from heapq import heapify, heappush, heappop
from vbweights import w
from comptab import fcomp
from bitcoding import B_dict
from inverse import inv
import itertools
from counters import counter
from glob import globs

if globs['split'] == 0: # splitting based on set of base relations
   from basesplit import bsplit as split
elif globs['split'] == 1: # splitting based on horn set
   from hornsplit import hsplit as split

DALL = B_dict['DALL']
cardmult = max(w)

# path consistency using van Beek weights
def iPCw(ConMatrix, neighbors, feeds):
   pq = []                         # list of entries arranged in a heap
   entry_finder = {}               # mapping of tasks to entries
   #REMOVED = None                  # placeholder for a removed task
   pcounter = itertools.count()     # unique sequence count
   Vars = len(ConMatrix)

   def add_task(task, priority=0):
       'Add a new task or update the priority of an existing task'
       if task in entry_finder:
           remove_task(task)
       count = next(pcounter)
       entry = [priority, count, task]
       entry_finder[task] = entry
       heappush(pq, entry)

   def remove_task(task):
       'Mark an existing task as REMOVED.  Raise KeyError if not found.'
       entry = entry_finder.pop(task)
       entry[-1] = None

   def pop_task():
       'Remove and return the lowest priority task. Raise KeyError if empty.'
       while pq:
           task = heappop(pq)[-1]
           if task:
               del entry_finder[task]
               return task
       raise KeyError('pop from an empty priority queue')   

   
   # initialize queue with weights and sort the queue
   map(add_task,[(m,n) for m, n in feeds if ConMatrix[m][n] != DALL], [(cardmult*split[ConMatrix[m][n]-1][0] + w[ConMatrix[m][n]-1]) for m, n in feeds if ConMatrix[m][n] != DALL])
   
   # as long as the queue is not empty, process it
   while pq:

      try:
         (i,j) = pop_task() # grab the appropriate relation
      except KeyError:
         break
      
      #counter['arcs'] += 1 # increment visited arcs counter 
  
      # create all triplets to be checked for path consistency
      for k in neighbors[i] & neighbors[j]:

          temp = fcomp[ConMatrix[k][i]-1][ConMatrix[i][j]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
  
             # constrain arc (k,i,j)
             temp = temp & ConMatrix[k][j]
             if temp != ConMatrix[k][j]:
                if not temp: 
                   return False # inconsistency
                ConMatrix[k][j] = temp
                ConMatrix[j][k] = inv[temp-1]
                if k < j:
                   add_task((k, j), cardmult*split[ConMatrix[k][j]-1][0] + w[ConMatrix[k][j]-1])
                else:
                   add_task((j, k), cardmult*split[ConMatrix[j][k]-1][0] + w[ConMatrix[j][k]-1])

          temp = fcomp[ConMatrix[i][j]-1][ConMatrix[j][k]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
          
             # constrain arc (i,j,k) 
             temp = temp & ConMatrix[i][k] 
             if temp != ConMatrix[i][k]:
                if not temp: 
                   return False # inconsistency
                ConMatrix[i][k] = temp
                ConMatrix[k][i] = inv[temp-1]
                if i < k:
                   add_task((i, k), cardmult*split[ConMatrix[i][k]-1][0] + w[ConMatrix[i][k]-1])
                else:
                   add_task((k, i), cardmult*split[ConMatrix[k][i]-1][0] + w[ConMatrix[k][i]-1])

   # the network is concistent and can't be refined further
   return True    
