from comptab import fcomp
from bitcoding import B_dict
from inverse import inv
from collections import deque
from counters import counter
from basesplit import bsplit as split
from pc import PC as pc
import sys

DALL = B_dict['DALL']

# path consistency using van Beek weights
def iSPC(ConMatrix, neighbors, feeds):
   pq = deque([])                         # list of entries arranged in a heap
   entry_finder = set([])               # mapping of tasks to entries
   Vars = len(ConMatrix)

   def add_task(task):
       'Add a new task or update the priority of an existing task'
       if task not in entry_finder:
          entry_finder.add(task)
          pq.append(task)
          
   def pop_task():
       'Remove and return the lowest priority task. Raise KeyError if empty.'
       task = pq.popleft()
       entry_finder.discard(task)
       return task

   # initialize queue 
   map(add_task,[(m,n) for m, n in feeds if ConMatrix[m][n] != DALL])

   # as long as the queue is not empty, process it
   while pq:

      (i,j) = pop_task() # grab the appropriate relation

      qcnchanges = {}
      
      mysplit = split[ConMatrix[i][j]-1][1]
 
      flag = True
      for b in mysplit:
          repairs = {}
          repairs[(i,j)] = ConMatrix[i][j]
          ConMatrix[i][j] = b
          ConMatrix[j][i] = inv[b-1]                  
          result = pc(ConMatrix, neighbors, i, j, repairs, [], False)
          if result:
             if flag:
                flag = False
                for c,v in repairs.keys():
                   qcnchanges[(c,v)] = ConMatrix[c][v]
             else:
                for c,v in repairs.keys():
                   if (c,v) in qcnchanges: 
                      qcnchanges[(c,v)] |= ConMatrix[c][v]   
                for c,v in qcnchanges.keys():
                   if (c,v) not in repairs:
                      del qcnchanges[(c,v)] 
          for g, h in repairs.keys():
             ConMatrix[g][h] = repairs[(g, h)] 
             ConMatrix[h][g] = inv[repairs[(g, h)]-1]

      if not qcnchanges:
         return False
     
      for g,h in qcnchanges.keys():
         if ConMatrix[g][h] != qcnchanges[(g,h)]:
            ConMatrix[g][h] = qcnchanges[(g,h)]
            ConMatrix[h][g] = inv[qcnchanges[(g,h)]-1]
            if (g,h) != (i,j):
               add_task((g, h)) 
          
   # the network is concistent and can't be refined further
   return True   
