from comptab import fcomp
from bitcoding import B_dict
from inverse import inv
from collections import deque
from counters import counter
from assignment import add_var, remove_var, mrv_variables
from itertools import combinations
from learned_weights import updateWeights
from glob import globs
from mapping import mapping

DALL = B_dict['DALL']

# path consistency using van Beek weights
def PC(ConMatrix, neighbors, m, n, repairs, trail, flag=True):
   pq = deque([])                         # list of entries arranged in a heap
   entry_finder = set([])               # mapping of tasks to entries
   Vars = len(ConMatrix)

   def add_task(task):
       'Add a new task or update the priority of an existing task'
       if task not in entry_finder:
          entry_finder.add(task)
          pq.append(task)
          
   def pop_task():
       'Remove and return the lowest priority task. Raise KeyError if empty.'
       task = pq.popleft()
       entry_finder.discard(task)
       return task

   add_task((m,n))

   # as long as the queue is not empty, process it
   while pq:

      (i,j) = pop_task() # grab the appropriate relation
      
      counter['arcs'] += 1 # increment visited arcs counter 
  
      # create all triplets to be checked for path consistency
      for k in neighbors[i] & neighbors[j]:

          temp = fcomp[ConMatrix[k][i]-1][ConMatrix[i][j]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
  
             # constrain arc (k,i,j)
             temp = temp & ConMatrix[k][j]
             if temp != ConMatrix[k][j]:
                if not temp: 
                   updateWeights(list(combinations([i,j,k],2)),counter['decisions'])
                   return False # inconsistency
                r = ConMatrix[k][j]
                ConMatrix[k][j] = temp
                ConMatrix[j][k] = inv[temp-1]
                if k < j:
                   add_task((k, j)) 
                   if (k, j) not in repairs:
                      repairs[(k, j)] = r
                   if flag:
                      if (k,j) not in mrv_variables:
                         if mapping[ConMatrix[k][j]-1] & ConMatrix[k][j] != mapping[ConMatrix[k][j]-1]:
                            add_var((k,j))
                            trail.append((1,(k,j)))
                      else:
                         if mapping[ConMatrix[k][j]-1] & ConMatrix[k][j] == mapping[ConMatrix[k][j]-1]:
                            remove_var((k,j))
                            trail.append((0,(k,j)))
                else:
                   add_task((j, k))     
                   if (j, k) not in repairs:
                      repairs[(j, k)] = inv[r-1]
                   if flag:
                      if (j,k) not in mrv_variables:
                         if mapping[ConMatrix[j][k]-1] & ConMatrix[j][k] != mapping[ConMatrix[j][k]-1]:
                            add_var((j,k))
                            trail.append((1,(j,k)))
                      else:
                         if mapping[ConMatrix[j][k]-1] & ConMatrix[j][k] == mapping[ConMatrix[j][k]-1]:
                            remove_var((j,k))
                            trail.append((0,(j,k)))

          temp = fcomp[ConMatrix[i][j]-1][ConMatrix[j][k]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
          
             # constrain arc (i,j,k) 
             temp = temp & ConMatrix[i][k] 
             if temp != ConMatrix[i][k]:
                if not temp: 
                   updateWeights(list(combinations([i,j,k],2)),counter['decisions'])
                   return False # inconsistency
                r = ConMatrix[i][k]
                ConMatrix[i][k] = temp
                ConMatrix[k][i] = inv[temp-1]
                if i < k:
                   add_task((i, k))                
                   if (i, k) not in repairs:
                      repairs[(i, k)] = r
                   if flag:
                      if (i,k) not in mrv_variables:
                         if mapping[ConMatrix[i][k]-1] & ConMatrix[i][k] != mapping[ConMatrix[i][k]-1]:
                            add_var((i,k))
                            trail.append((1,(i,k)))
                      else:
                         if mapping[ConMatrix[i][k]-1] & ConMatrix[i][k] == mapping[ConMatrix[i][k]-1]:
                            remove_var((i,k))
                            trail.append((0,(i,k)))
                else:
                   add_task((k, i))                  
                   if (k, i) not in repairs:
                      repairs[(k, i)] = inv[r-1]
                   if flag:
                      if (k,i) not in mrv_variables:
                         if mapping[ConMatrix[k][i]-1] & ConMatrix[k][i] != mapping[ConMatrix[k][i]-1]:
                            add_var((k,i))
                            trail.append((1,(k,i)))
                      else:
                         if mapping[ConMatrix[k][i]-1] & ConMatrix[k][i] == mapping[ConMatrix[k][i]-1]:
                            remove_var((k,i))
                            trail.append((0,(k,i))) 

   # the network is concistent and can't be refined further
   return True    
